@@ -23,12 +23,11 @@
 
 namespace Hyprutils::Memory {
     namespace Atomic_ {
-        template <typename T>
-        class impl : public Impl_::impl<T> {
+        class impl : public Impl_::impl_base {
             std::recursive_mutex m_mutex;
 
           public:
-            impl(T* data, bool lock = true) noexcept : Impl_::impl<T>(data, lock) {
+            impl(void* data, DeleteFn deleter) noexcept : Impl_::impl_base(data, deleter) {
                 ;
             }
 
@@ -55,7 +54,13 @@ namespace Hyprutils::Memory {
         using validHierarchy = std::enable_if_t<std::is_assignable_v<CAtomicSharedPointer<T>&, X>, CAtomicSharedPointer&>;
 
       public:
-        explicit CAtomicSharedPointer(Impl_::impl_base* impl) noexcept : m_ptr(impl) {}
+        explicit CAtomicSharedPointer(T* object) noexcept : m_ptr(new Atomic_::impl(sc<void*>(object), _delete), sc<void*>(object)) {
+            ;
+        }
+
+        CAtomicSharedPointer(Impl_::impl_base* impl, void* data) noexcept : m_ptr(impl, data) {
+            ;
+        }
 
         CAtomicSharedPointer(const CAtomicSharedPointer<T>& ref) {
             if (!ref.m_ptr.impl_)
@@ -141,7 +146,7 @@ namespace Hyprutils::Memory {
             // -> must unlock BEFORE reset
             // not last ref?
             // -> must unlock AFTER reset
-            auto& mutex = sc<Atomic_::impl<T>*>(m_ptr.impl_)->getMutex();
+            auto& mutex = sc<Atomic_::impl*>(m_ptr.impl_)->getMutex();
             mutex.lock();
 
             if (m_ptr.impl_->ref() > 1) {
@@ -152,7 +157,11 @@ namespace Hyprutils::Memory {
 
             if (m_ptr.impl_->wref() == 0) {
                 mutex.unlock(); // Don't hold the mutex when destroying it
-                m_ptr.reset();
+
+                m_ptr.impl_->destroy();
+                delete sc<Atomic_::impl*>(m_ptr.impl_);
+                m_ptr.impl_ = nullptr;
+
                 // mutex invalid
                 return;
             } else {
@@ -163,12 +172,18 @@ namespace Hyprutils::Memory {
                 // To avoid this altogether, keep a weak pointer here.
                 // This guarantees that impl_ is still valid after the reset.
                 CWeakPointer<T> guard = m_ptr;
-                m_ptr.reset();
+                m_ptr.reset(); // destroys the data
 
                 // Now we can safely check if guard is the last wref.
                 if (guard.impl_->wref() == 1) {
                     mutex.unlock();
-                    return; // ~guard destroys impl_ and mutex
+
+                    // destroy impl_ (includes the mutex)
+                    delete sc<Atomic_::impl*>(guard.impl_);
+                    guard.impl_ = nullptr;
+
+                    // mutex invalid
+                    return;
                 }
 
                 guard.reset();
@@ -204,9 +219,17 @@ namespace Hyprutils::Memory {
             return m_ptr.impl_ ? m_ptr.impl_->ref() : 0;
         }
 
+        Atomic_::impl* impl() const {
+            return sc<Atomic_::impl*>(m_ptr.impl_);
+        }
+
       private:
+        static void _delete(void* p) {
+            std::default_delete<T>{}(sc<T*>(p));
+        }
+
         std::lock_guard<std::recursive_mutex> implLockGuard() const {
-            return sc<Atomic_::impl<T>*>(m_ptr.impl_)->lockGuard();
+            return impl()->lockGuard();
         }
 
         CSharedPointer<T> m_ptr;
@@ -312,11 +335,13 @@ namespace Hyprutils::Memory {
             // -> must unlock BEFORE reset
             // not last ref?
             // -> must unlock AFTER reset
-            auto& mutex = sc<Atomic_::impl<T>*>(m_ptr.impl_)->getMutex();
+            auto& mutex = sc<Atomic_::impl*>(m_ptr.impl_)->getMutex();
             mutex.lock();
             if (m_ptr.impl_->ref() == 0 && m_ptr.impl_->wref() == 1) {
                 mutex.unlock();
-                m_ptr.reset();
+
+                delete sc<Atomic_::impl*>(m_ptr.impl_);
+                m_ptr.impl_ = nullptr;
                 // mutex invalid
                 return;
             }
@@ -370,12 +395,16 @@ namespace Hyprutils::Memory {
             if (!m_ptr.impl_->dataNonNull() || m_ptr.impl_->destroying() || !m_ptr.impl_->lockable())
                 return {};
 
-            return CAtomicSharedPointer<T>(m_ptr.impl_);
+            return CAtomicSharedPointer<T>(m_ptr.impl_, m_ptr.m_data);
+        }
+
+        Atomic_::impl* impl() const {
+            return sc<Atomic_::impl*>(m_ptr.impl_);
         }
 
       private:
         std::lock_guard<std::recursive_mutex> implLockGuard() const {
-            return sc<Atomic_::impl<T>*>(m_ptr.impl_)->lockGuard();
+            return impl()->lockGuard();
         }
 
         CWeakPointer<T> m_ptr;
@@ -387,7 +416,22 @@ namespace Hyprutils::Memory {
     };
 
     template <typename U, typename... Args>
-    static CAtomicSharedPointer<U> makeAtomicShared(Args&&... args) {
-        return CAtomicSharedPointer<U>(new Atomic_::impl<U>(new U(std::forward<Args>(args)...)));
+    [[nodiscard]] inline CAtomicSharedPointer<U> makeAtomicShared(Args&&... args) {
+        return CAtomicSharedPointer<U>(new U(std::forward<Args>(args)...));
+    }
+
+    template <typename T, typename U>
+    CAtomicSharedPointer<T> reinterpretPointerCast(const CAtomicSharedPointer<U>& ref) {
+        return CAtomicSharedPointer<T>(ref.impl(), ref.m_data);
+    }
+
+    template <typename T, typename U>
+    CAtomicSharedPointer<T> dynamicPointerCast(const CAtomicSharedPointer<U>& ref) {
+        if (!ref)
+            return nullptr;
+        T* newPtr = dynamic_cast<T*>(sc<U*>(ref.impl()->getData()));
+        if (!newPtr)
+            return nullptr;
+        return CAtomicSharedPointer<T>(ref.impl(), newPtr);
     }
 }
